import logging
from logging import handlers as log_handlers

import flask

from lpreports import auth
from lpreports import db
from lpreports import rest_api
from lpreports import settings
from lpreports.views import about as about_blueprint
from lpreports.views import analytics as analytics_blueprint
from lpreports.views import charts as charts_blueprint
from lpreports.views import cve as cve_blueprint
from lpreports.views import login as login_blueprint
from lpreports.views import reports as reports_blueprint
from lpreports.views import template_helpers


log = logging.getLogger(__name__)


def load_reviews_config(app):
    return app.config.load_config("review.json")


def register_error_handler(app, code, page):
    app.register_error_handler(
        code,
        lambda error: (flask.render_template(page), code))


# TODO(ruhe): log to console if config != Production
def setup_logging(app):
    file_handler = log_handlers.RotatingFileHandler(
        "/var/log/lpreports/lpreports.log",
        maxBytes=1024 * 1024 * 100,
        backupCount=20)

    file_handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "[%(asctime)s] - %(name)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(formatter)

    app.logger.addHandler(file_handler)
    logging.getLogger("requests").setLevel(logging.ERROR)
    logging.getLogger("iso8601").setLevel(logging.ERROR)


def create_app(test_mode=False, enable_logging=True):
    app = flask.Flask(__name__)
    if not test_mode and enable_logging:
        setup_logging(app)

    log.info("Configuring application")
    # load default settings
    app.config.from_object(settings.Default())
    # load test or non-test settings
    if test_mode:
        app.config.from_object(settings.Testing())
    else:
        app.config.from_object(settings.Production())
    db.init_app(app)
    auth.init_app(app)

    template_helpers.register_jinja_globals(app,
                                            app.config['REPORT_CONFIG'],
                                            app.config['REVIEWS'],
                                            app.config['KEY_MILESTONES'])

    app.register_blueprint(auth.auth, url_prefix='/auth')
    app.register_blueprint(about_blueprint.about_page, url_prefix='/')
    app.register_blueprint(analytics_blueprint.analytics,
                           url_prefix='/analytics')
    app.register_blueprint(charts_blueprint.charts, url_prefix='/charts')
    app.register_blueprint(cve_blueprint.cve, url_prefix='/cve')
    app.register_blueprint(login_blueprint.login_page, url_prefix='/login')
    app.register_blueprint(reports_blueprint.reports, url_prefix='/reports')
    app.register_blueprint(rest_api.api_bp, url_prefix='/api/v1.0')

    app.before_request(auth.requires_auth)

    register_error_handler(app, 401, "401.html")
    register_error_handler(app, 404, "404.html")
    register_error_handler(app, 500, "500.html")

    return app
