import collections
import json
import logging
import os
import yaml

from flask import current_app as app

from lpreports import db
from lpreports.exceptions import ImproperlyConfigured
from lpreports.launchpad import client as lp_client
from lpreports.launchpad import exceptions

log = logging.getLogger(__name__)


def ordered_load(stream):
    class OrderedLoader(yaml.Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return collections.OrderedDict(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)

    return yaml.load(stream, OrderedLoader)


def load_config(config_name):
    for conf_dir in app.config['CONF_DIRS']:
        path = os.path.join(conf_dir, config_name)
        if os.path.exists(path):
            with open(path) as fl:
                return ordered_load(fl)

    raise ImproperlyConfigured('%s not found in %r' % (config_name, app.config['CONF_DIRS']))


def process_staff_line(subteam, conf_staff_line, lp_client):
    if 'include-lp-team' in conf_staff_line:
        team_name = conf_staff_line['include-lp-team']
        try:
            members = lp_client.get_team_members(team_name)
        except exceptions.ClientError:
            log.error("Members of '{0}' haven't been "
                      "received. (Import was skipped).".format(team_name))
        else:
            subteam.add(team_name)
            for member in members:
                subteam.add(member["name"])
    elif 'include-lp-person' in conf_staff_line:
        person = conf_staff_line['include-lp-person']
        subteam.add(person)
    elif 'exclude-lp-person' in conf_staff_line:
        exclude = conf_staff_line['exclude-lp-person']
        subteam.discard(exclude)
    else:
        raise RuntimeError('Can not parse instruction: "%s"' % conf_staff_line)


def process_teams_config(teams_config, lp_client):
    teams = collections.OrderedDict()
    for conf_team_name in teams_config.keys():
        log.info(">>> Importing team {0}".format(conf_team_name))

        team = {}

        for conf_subteam in teams_config[conf_team_name]:
            subteam_name = conf_subteam.keys()[0]
            log.info("Importing subteam {0}".format(subteam_name))
            subteam = set()

            for conf_staff_line in conf_subteam.values()[0]:
                process_staff_line(subteam, conf_staff_line, lp_client)

            team[subteam_name] = list(subteam)

        teams[conf_team_name] = team

    return teams


def collect_assignees():
    lp = lp_client.LaunchpadClient(app.config["LP_CONSUMER_KEY"],
                                   app.config["LP_TOKEN"],
                                   app.config["LP_SIGNATURE"])

    teams_config = load_config('teams.yaml')
    teams = process_teams_config(teams_config['teams'], lp)

    assignees = db.assignees_db()
    major_teams = assignees.major_teams
    assignees.drop_collection(major_teams)

    for team_name, subteams in teams.items():
        people = set()
        for subteam in subteams.values():
            people.update(subteam)

        major_teams.insert({'Team': team_name,
                            'Members': list(people)})

    full_teams = assignees.full_teams
    assignees.drop_collection(full_teams)
    full_teams.insert({'Full teams': json.dumps(teams)})


def collect_security_teams():
    lp = lp_client.LaunchpadClient(app.config["LP_CONSUMER_KEY"],
                                   app.config["LP_TOKEN"],
                                   app.config["LP_SIGNATURE"])

    teams_config = load_config('teams.yaml')
    security_teams = teams_config["security-teams"]

    log.info("Collecting members of security teams: {0}".format(
        str(security_teams)))

    users = collections.defaultdict(list)

    for team in security_teams:
        target = team.split("-")[0]
        members = lp.get_team_members(team)

        for user in members:
            email = user["email"]
            if email:
                users[email].append(target)
            else:
                log.warning("User {0} doesn't have email associated "
                            "with account".format(user["display_name"]))

    users = [{"email": _email, "projects": _projects} for _email, _projects
             in users.items()]

    security_team_db = db.security_team_db()
    people = security_team_db.people
    security_team_db.drop_collection(people)
    people.insert(users)


def load_area_tag_teams():
    area_tag_teams = load_config('teams.yaml')['area-tag-teams']

    area_tags_db = db.area_tags_db()
    area_tags_db.drop_collection(area_tags_db.teams)

    log.info("Loading area-tag teams")
    for team_name, subteams in area_tag_teams.items():
        record = dict(name=team_name, subteams=[])
        for subteam_name, subteam in subteams.items():
            tags = subteam.get('area-tags')
            if tags:
                record['subteams'].append(
                    dict(name=subteam_name, area_tags=tags)
                )
        if record['subteams']:
            area_tags_db.teams.insert(record)


def sync_teams():
    collect_assignees()
    collect_security_teams()
    load_area_tag_teams()
