import logging
import os
import yaml

from git import Repo

from lpreports import db

log = logging.getLogger(__name__)

REMOTE_URL = "git://git.openstack.org/openstack/ossa"
LOCAL_REPO_PATH = "/tmp/ossa"


def fetch_ossa_repo():
    log.info("Fetching from {0}".format(REMOTE_URL))

    repo = Repo.init(LOCAL_REPO_PATH)
    if not repo.remotes:
        origin = repo.create_remote("origin", REMOTE_URL)
    else:
        origin = repo.remotes[0]
    origin.fetch()
    origin.pull(origin.refs[0].remote_head)

    log.info("Fetched data from OSSA repo")


def scan_ossa_dir(path):
    for f in os.listdir(path):
        yield os.path.join(path, f)


def find_bugs_with_cve(bugs, cve):
    for bug in bugs:
        if cve in bug.get("cves", []):
            yield "https://launchpad.net/bugs/{0}".format(bug["id"])


def parse_ossa(path):
    all_bugs = db.get_all_bugs()

    with open(path, "r") as stream:
        raw = yaml.load(stream)
        cves = [x["cve-id"] for x in raw["vulnerabilities"]]

        related_bugs = []
        for c in cves:
            for f in find_bugs_with_cve(all_bugs, c):
                related_bugs.append(f)

        return {
            "ossa_id": raw["id"],
            "ossa_date": str(raw["date"]),
            "title": raw["title"],
            "affected_products": [x["product"] for x in raw["affected-products"]],
            "cves": cves,
            "upstream_bugs": raw["issues"]["links"],
            "product_bugs": list(set(related_bugs))
        }


def generate_report():
    fetch_ossa_repo()
    vulnerabilities = []

    for f in scan_ossa_dir(LOCAL_REPO_PATH + "/ossa"):
        vulnerabilities.append(parse_ossa(f))

    return reversed(vulnerabilities)


def update_cves():
    log.info("Generating CVE report")
    report = generate_report()
    log.info("Finished generating CVE report")

    cve_db = db.cve_db()
    cves = cve_db.cves
    cve_db.drop_collection(cves)
    cves.insert(report)
