import logging
import sys
import threading
from time import sleep

from flask import current_app as app

from lpreports import db
from lpreports.launchpad import client as lpclient
from lpreports.launchpad import exceptions

log = logging.getLogger(__name__)


class SynchronizerThread(threading.Thread):
    def __init__(self, buglist, lp, projects, bugs_db, *args, **kwargs):
        self.lp = lp
        self.projects = projects
        self.buglist = buglist
        self.bugs_db = bugs_db
        self._stop = threading.Event()
        super(self.__class__, self).__init__(*args, **kwargs)

    def stop(self):
        self._stop.set()

    def is_stopped(self):
        return self._stop.isSet()

    def run(self):
        processed_bugs_count = 0
        for i in range(100):  # process 100 bugs per thread
            # kill the thread if it is stopped
            if self.is_stopped():
                return

            # kill the thread if bugs list is empty
            try:
                bug_id = self.buglist.pop()
                processed_bugs_count += 1
            except (IndexError, KeyError):
                break

            processed_tasks = []
            try:
                bug_tasks = self.lp.get_bugtasks_by_id(bug_id)
            except exceptions.NotFound:
                log.warn("Bug #%s wasn't found so it will be skipped.", bug_id)
                continue
            except Exception:
                # on any failure add bug back to list and forward to next bug
                log.error("Failed processing bug #%s", bug_id)
                self.buglist.append(bug_id)
                continue

            for task in bug_tasks:
                processed_task = self.lp.process_bug_task(task)

                # skip unknown projects
                if processed_task['target_name'] not in self.projects:
                    continue

                log.info('Updating bug #{id} - {title} in [{project}: {milestone}]'.format(
                    id=processed_task['id'],
                    title=processed_task['title'].encode('ascii', 'ignore'),
                    project=processed_task['target_name'],
                    milestone=processed_task['milestone']
                ))
                processed_tasks.append(processed_task)

            db.update_bugs(processed_tasks, self.bugs_db)
        log.debug('Processed %s bugs' % processed_bugs_count)


class MilestoneSynchronizer(object):
    def __init__(self, lp, projects):
        self.lp = lp
        self.projects = projects
        self.local_etags = dict(db.get_milestones_etag())
        self.remote_etags = {p: lp.get_milestone_etags(p) for p in projects}
        self.changed_milestones = self._find_changed_milestones()
        self.stale_bugs = self._find_stale_bugs()

    def _find_changed_milestones(self):
        changed_milestones = {}
        for project in self.projects:
            # get difference between local and remote milestones
            local = self.local_etags.get(project, {})
            remote = self.remote_etags.get(project, {})
            changed = list(set(local) ^ set(remote))

            # found milestones which don't have the same etag
            for key in set(local) & set(remote):
                if local[key] != remote[key]:
                    changed.append(key)

            changed_milestones[project] = changed
        return changed_milestones

    def _find_stale_bugs(self):
        stale_bugs = []
        for project, milestones in self.changed_milestones.items():
            if milestones:
                bugs = db.get_bugs(project, milestone_names=milestones)
                stale_bugs.extend([b.id for b in bugs])
        return list(set(stale_bugs))

    def update_local_etags(self):
        """Save remote etags of milestones as local"""
        db.update_milestones_etag(self.remote_etags)

    def get_changed_milestones(self):
        """Get changed milestones as dict {project_name: list_of_milestones, ...}"""
        return self.changed_milestones

    def get_stale_bugs(self):
        """Get bugs which should be resynchronized"""
        return self.stale_bugs


def run_sync(num_threads=5):
    # don't run repair job if database is empty
    if db.get_update_timestamp() == 0:
        log.warn("Database is empty so repair job won't be started.")
        sys.exit(2)

    lp = lpclient.LaunchpadClient(
        app.config["LP_CONSUMER_KEY"],
        app.config["LP_TOKEN"],
        app.config["LP_SIGNATURE"]
    )
    projects = app.config['PROJECTS_LIST']

    msync = MilestoneSynchronizer(lp, projects)

    log.debug('Changed milestones grouped by project :%r', msync.get_changed_milestones())
    log.debug('Count of bugs to update: %d', len(msync.get_stale_bugs()))

    thread_list = []

    try:
        while True:
            bugs = msync.get_stale_bugs()
            thread_count = len(thread_list)

            if not bugs and not thread_list:
                log.info("Finished processing bugs")
                break
            log.debug("Remaining bugs: %s, remaining threads %s", len(bugs), thread_count)
            if thread_count < num_threads:
                for i in range(thread_count, min(num_threads, len(bugs))):
                    thread = SynchronizerThread(bugs, lp, projects, db.bugs_db())
                    thread_list.append(thread)
                    thread.start()
            sleep(1)  # simple timeout for while loop
            for thread in thread_list:
                if not thread.is_alive():
                    thread_list.remove(thread)
    except KeyboardInterrupt:
        log.info('Ctrl-c invoked! Stopping all threads...')
        for thread in thread_list:
            thread.stop()
            thread.join()
        sys.exit(1)

    msync.update_local_etags()
