import functools
import logging
import operator
import Queue
import threading
import time

from flask import current_app as app

from lpreports import db
from lpreports.launchpad import client as lp_client

log = logging.getLogger(__name__)

NUM_THREADS = 2


class ProducerThread(threading.Thread):
    def __init__(self, queue, stop_event, lp, project, modified_since):
        super(ProducerThread, self).__init__()
        self.queue = queue
        self.stop_event = stop_event
        self.lp = lp
        self.project = project
        self.modified_since = modified_since

    def run(self):
        try:
            for bugs in self.lp.find_bug_tasks(self.project, self.modified_since):
                self.queue.put(bugs)
        except Exception as e:
            log.error(e, exc_info=True)
        finally:
            log.info("Producer finished produsing "
                     "bug tasks for {0}".format(self.project))
            self.stop_event.set()


class ConsumerThread(threading.Thread):
    def __init__(self, queue, stop_event, lp, project, bugs_db):
        super(ConsumerThread, self).__init__()
        self.queue = queue
        self.stop_event = stop_event
        self.lp = lp
        self.project = project
        self.bugs_db = bugs_db
        self.projects_list = app.config["PROJECTS_LIST"]

    def run(self):
        while True:
            try:
                bugs = self.queue.get(block=False)
                try:
                    processed_bugs = [self.lp.process_bug_task(b) for b in bugs]

                    filtered_bugs = list()
                    for processed_bug in processed_bugs:
                        target_name = processed_bug["target_name"]
                        if target_name in self.projects_list:
                            log.info("Adding bug #{0} - {1} to [{2} : {3}]".format(
                                processed_bug["id"],
                                processed_bug["title"].encode("ascii", "ignore"),
                                processed_bug["target_name"],
                                processed_bug["milestone"]))
                            filtered_bugs.append(processed_bug)

                    db.update_bugs(filtered_bugs, self.bugs_db)
                except Exception as e:
                    log.error(e, exc_info=True)
                    log.error(str(bugs))
            except Queue.Empty:
                if self.stop_event.is_set():
                    log.info("Consumer finished processing "
                             "bugs for {0}".format(self.project))
                    break
                time.sleep(0.2)


def fetch_bugs(modified_since):
    lp = lp_client.LaunchpadClient(app.config["LP_CONSUMER_KEY"],
                                   app.config["LP_TOKEN"],
                                   app.config["LP_SIGNATURE"])

    producers = []
    consumers = []

    for project in app.config["PROJECTS_LIST"]:
        queue = Queue.Queue()
        stop_event = threading.Event()

        if not db.bugs_db()[project].count():
            modified_since = False

        producer = ProducerThread(queue, stop_event, lp,
                                  project, modified_since)
        producers.append(producer)

        for i in range(NUM_THREADS + 1):
            consumer = ConsumerThread(queue, stop_event,
                                      lp, project,
                                      db.bugs_db())
            consumers.append(consumer)

    for producer in producers:
        producer.start()

    for consumer in consumers:
        consumer.start()

    for consumer in consumers:
        consumer.join()


def syncdb(drop_db=False):
    sync_start_time = time.time()

    lp = lp_client.LaunchpadClient(app.config["LP_CONSUMER_KEY"],
                                   app.config["LP_TOKEN"],
                                   app.config["LP_SIGNATURE"])
    projects = app.config['PROJECTS_LIST']

    if drop_db:
        for project in projects:
            log.info('Removing %s collection', project)
            getattr(db.bugs_db(), project).drop()

    milestones_list = functools.reduce(operator.or_, map(
        lambda p: set(lp.get_project(p)["all_milestones"]), projects))

    active_milestones_list = functools.reduce(operator.or_, map(
        lambda p: set(lp.get_project(p)["active_milestones"]), projects))

    active_milestones_target_dates = {}
    for project in projects:
        if project in app.config["DEFAULT_PROJECTS"]:
            ms = lp.get_project(project)["active_milestones_target_dates"]
            active_milestones_target_dates.update(ms)

    log.info("Creating milestones...")
    db.upsert_milestones(milestones_list,
                         active_milestones_list,
                         active_milestones_target_dates)

    fetch_bugs(db.get_update_time())

    db.set_update_time(sync_start_time)

    lp.close_session()
