import collections
import json
import logging

import cachetools.func
from flask import current_app as app
from flask.ext.pymongo import PyMongo

from lpreports.launchpad.bug import Bug
from lpreports.launchpad.milestone import Milestone
import lpreports.utils.common

log = logging.getLogger(__name__)

bugs = None
assignees = None
area_tags = None
users = None
security = None
cve = None


def init_app(app):
    global bugs, assignees, area_tags, users, security, cve
    bugs = PyMongo(app, config_prefix='MONGO_BUGS')
    assignees = PyMongo(app, config_prefix='MONGO_ASSIGNEES')
    area_tags = PyMongo(app, config_prefix='MONGO_AREA_TAGS')
    users = PyMongo(app, config_prefix='MONGO_USERS')
    security = PyMongo(app, config_prefix='MONGO_SECURITY')
    cve = PyMongo(app, config_prefix='MONGO_CVE')


def bugs_db():
    return bugs.db


def assignees_db():
    return assignees.db


def area_tags_db():
    return area_tags.db


def users_db():
    return users.db


def security_team_db():
    return security.db


def cve_db():
    return cve.db


def get_update_timestamp():
    try:
        ts = int(bugs_db().update_date.find_one()["Update_date"])
    except Exception:
        ts = 0
        log.warn("Failed to get last update time. "
                 "Defaulting to {0}".format(ts))

    return ts


def get_update_time():
    ts = get_update_timestamp()
    return lpreports.utils.common.timestamp_to_utc_date(ts)


def set_update_time(update_time):
    bugs_db().drop_collection("update_date")
    bugs_db().create_collection("update_date")
    bugs_db().update_date.insert({"Update_date": update_time})


def find_user(**kwargs):
    store = users_db().users
    return store.find_one(kwargs)


def add_user(user, update=False):
    store = users_db().users

    if update:
        store.update({"_id": user["_id"]}, user)
    else:
        store.insert(user)


def get_user_security_teams_membership(email):
    store = security_team_db().people
    user = store.find_one({"email": email})

    if user:
        return user["projects"]
    else:
        return []


def get_milestones(active_only=False):
    id = 2 if active_only else 1
    raw = bugs_db().milestones.find_one({"_id": id})["Milestone"]
    major = [Milestone(m) for m in raw if "-" not in m]
    updates = [Milestone(m) for m in raw if "-" in m]

    ordered = []
    for lst in (major, updates):
        ordered.extend(sorted(lst, reverse=True))
    return ordered


def get_maintenance_milestones():
    return [ms for ms in get_milestones(active_only=True)
            if ms not in app.config['KEY_MILESTONES']]


def get_milestones_target_dates():
    return bugs_db().milestones.find_one({"_id": 3})["Milestone"]


def upsert_milestones(milestones, active_milestones,
                      active_milestones_target_dates):
    bugs_db().milestones.remove({"_id": 1})
    bugs_db().milestones.insert(
        {"_id": 1, "Milestone": sorted(milestones, reverse=True)})
    bugs_db().milestones.remove({"_id": 2})
    bugs_db().milestones.insert(
        {"_id": 2, "Milestone": sorted(active_milestones, reverse=True)})
    bugs_db().milestones.remove({"_id": 3})
    bugs_db().milestones.insert(
        {"_id": 3, "Milestone": active_milestones_target_dates},
        check_keys=False)


def get_major_teams():
    teams = collections.OrderedDict()

    for mteam in assignees_db().major_teams.find():
        teams[mteam['Team']] = mteam['Members']

    teams['Unknown'] = {'unknown': []}

    return teams


@cachetools.func.ttl_cache(ttl=10)
def get_full_teams():
    raw_data = assignees_db().full_teams.find_one()['Full teams']
    teams_data = json.loads(raw_data,
                            object_pairs_hook=collections.OrderedDict)
    teams_data['Unknown'] = {'unknown': []}
    return teams_data


def get_teams():
    full_teams = get_full_teams()
    result = collections.OrderedDict()
    for department, teams in full_teams.items():
        result[department] = sorted(teams.keys())
    return result


def get_department_members(department):
    """Get all members of given Mirantis department."""
    return assignees_db().major_teams.find_one({"Team": department})['Members']


def update_bugs(bugs_lst, _bugs_db):
    bugs_by_target = dict()
    for bug in bugs_lst:
        bugs_by_target.setdefault(bug["target_name"], list()).append(bug)

    for target, bugs in bugs_by_target.items():
        _bugs_db[target].delete_many({'id': bugs[0]['id']})
        for bug in bugs:
            _bugs_db[target].insert(bug)


def delete_bug_task(project, task):
    bugs_db()[project].remove(task)


@cachetools.func.ttl_cache()
def get_all_bugs():
    # should be deprecated, because each tool must have its own project list
    # configuration. DEFAULT_PROJECTS sets default filter value.
    return sum(map(lambda p: get_all_bugs_in_project(p),
                   app.config['DEFAULT_PROJECTS']), [])


@cachetools.func.ttl_cache()
def get_all_bugs_in_project(project):
    return [bug for bug in bugs_db()[project].find()]


def get_bugs(project_name, statuses=None, milestone_names=None,
             tags=None, importance=None, condition=False, **kwargs):
    """Get bugs from database for project

    :param project_name: Project name, required
    :param statuses: statuses list, default: []
    :param milestone_names: milestone names list, default: []
    :param tags: tags list, default: []
    :param importance: importance values list, default: []
    :param condition: tags search flag: True => exclude from search, default: False
    :return: bugs list
    """

    project = bugs_db()[project_name]

    search = []

    if statuses:
        search = [{"status": {"$in": statuses}}]
    if milestone_names:
        search.append({'milestone': {"$in": milestone_names}})
    if importance:
        search.append({"importance": {"$in": importance}})

    if tags:
        condition = "$nin" if condition else "$in"
        search.append({"tags": {condition: tags}})

    if len(search) == 1:
        query = search[0]  # simple search
    elif len(search) > 1:
        query = {"$and": search}  # complex search
    else:
        query = {}  # find_all

    # return generator
    return map(lambda x: Bug(x), project.find(query))


def get_cves():
    return cve_db().cves.find()


def get_area_tag_teams():
    result = collections.OrderedDict()
    for team in area_tags_db().teams.find():
        result[team['name']] = [x['name'] for x in team['subteams']]
    return result


def get_area_tags(team):
    for doc in area_tags_db().teams.find({"subteams.name": {"$in": [team]}}):
        for subteam in doc["subteams"]:
            if subteam["name"] == team:
                return subteam["area_tags"]
    return None


def update_milestones_etag(milestones):
    bugs_db().milestone_etag.drop()
    for project in milestones:
        milestone_list = [dict(name=m, etag=e) for m, e in milestones[project].items()]
        bugs_db().milestone_etag.insert(
            {'project': project, 'milestones': milestone_list}
        )


def get_milestones_etag():
    result = {}
    for record in bugs_db().milestone_etag.find():
        project = record['project']
        milestones = {m['name']: m['etag'] for m in record['milestones']}
        result[project] = milestones
    return result
