# This is a wrapper, which converts launchpad bug structure into
# a internal bug object.
#
# When you get certain properties of the bug (e.g. assignee),
# it usually does additional query to LP.
# This wrapper avoids doing any calls to Launchpad by going into internal
# representation of the object and grabbing the info from JSON.


FIELDS_TO_COPY = [
    "date_assigned",
    "date_closed",
    "date_confirmed",
    "date_created",
    "date_fix_committed",
    "date_fix_released",
    "date_in_progress",
    "date_incomplete",
    "date_left_closed",
    "date_left_new",
    "date_triaged",
    "date_last_updated",
    "id",
    "web_link",
    "milestone",
    "milestone_link",
    "status",
    "tags",
    "target_name",
    "title",
    "importance",
    "owner",
    "owner_link",
    "assignee",
    "assignee_link",
    "verifier",
    "private",
    "information_type",
    "is_duplicate"
]

BUG_STATUSES = {
    "New": ["New"],
    "Incomplete": ["Incomplete"],
    "Open": ["Confirmed", "Triaged"],
    "In Progress": ["In Progress"],
    "Resolved": ["Fix Committed", "Won't Fix", "Invalid", "Opinion"],
    "Verified": ["Fix Released"]
}


class Bug(object):
    def __init__(self, lpbug):

        # straight copy fields from the lpbug object. this do not
        # make any calls to LP
        for name in FIELDS_TO_COPY:
            setattr(self, name, lpbug.get(name))
        self.web_link = "https://launchpad.net/bugs/{0}".format(self.id)
        self.assignee_link = "http://launchpad.net/~{0}".format(self.assignee)
        self.weight = None
        self.bug_tasks = []

    def get_status_changes(self):
        # Bug statuses:
        # * Incomplete -> date_incomplete
        # * New (not targeted to any release)  ->  date_created
        # * Open        -> date_triaged (date_confirmed, date_left_new,
        # date_assigned)
        # * In Progress -> date_in_progress
        # * Resolved    -> date_fix_committed
        # * Verified    -> date_fix_released

        # if the bug is "New", it should not be displayed on the chart
        if self.status in BUG_STATUSES["New"]:
            return []

        # list of dates
        result = []

        # When the bug was assigned to the release
        date_open = min(d for d in [self.date_triaged, self.date_confirmed,
                                    self.date_left_new, self.date_assigned]
                        if d is not None)
        result.append({
            "date": date_open,
            "type": "Open",
            "matches": BUG_STATUSES["Open"]
        })

        # When the bug went to in progress state
        date_in_progress = self.date_in_progress
        result.append({
            "date": date_in_progress,
            "type": "In Progress",
            "matches": BUG_STATUSES["In Progress"]
        })

        # When the bug was resolved or closed (e.g. as invalid)
        date_resolved = next((d for d in [self.date_fix_committed,
                                          self.date_closed]
                              if d is not None), None)
        result.append({
            "date": date_resolved,
            "type": "Resolved",
            "matches": BUG_STATUSES["Resolved"]
        })

        # When the bug was verified
        date_verified = self.date_fix_released
        result.append({
            "date": date_verified,
            "type": "Verified",
            "matches": BUG_STATUSES["Verified"]
        })

        # When the bug was set as incomplete
        date_incomplete = self.date_incomplete
        result.append({
            "date": date_incomplete,
            "type": "Incomplete",
            "matches": BUG_STATUSES["Incomplete"]})

        # Remove all entries which have date as "None"
        result = [e for e in result if e["date"] is not None]

        # Filter dates and statuses which are out of line
        for i in range(0, len(result)):
            for j in range(i + 1, len(result)):
                if result[i]["date"] > result[j]["date"]:
                    result[i]["obsolete"] = True

        # Remove all obsoleted entries
        result = [e for e in result if "obsolete" not in e]

        # Find the first element which matches our bug status
        idx = -1
        for i in range(0, len(result)):
            if self.status in result[i]["matches"]:
                idx = i
                break

        # The date for our status is not found. Not sure if it can happen
        if idx < 0:
            return []

        # Get the corresponding prefix of result,
        # so it ends with the right status (the status in which our bug is in)
        result = result[:idx + 1]

        return result

    def set_weight(self, weight):
        self.weight = weight

    def set_legend(self, legend):
        self._legend = legend

    @property
    def major_team(self):
        """Returns first matching major team name."""
        from lpreports import db
        teams_data = db.get_full_teams()

        for major_team_name, subteam in teams_data.iteritems():
            for subteam_name, subteam_members in subteam.iteritems():
                if self.assignee in subteam_members:
                    return major_team_name
        return 'Unknown'

    @property
    def legend(self,):
        legend = []
        for weight, description in self._legend:
            legend.append("{0}: {1}".format(weight, description))
        return "\n".join(legend)
