import datetime

import abc
import pytz

from lpreports import db
from lpreports.launchpad import filters

CRITICAL_IMPORTANCE = "Critical"
HIGH_IMPORTANCE = "High"

NEW_STATUS = "New"
INCOMPLETE_STATUS = "Incomplete"
OPINION_STATUS = "Opinion"
INVALID_STATUS = "Invalid"
WONT_FIX_STATUS = "Won't Fix"
CONFIRMED_STATUS = "Confirmed"
TRIAGED_STATUS = "Triaged"
IN_PROGRESS_STATUS = "In Progress"
FIX_COMMITTED_STATUS = "Fix Committed"
FIX_RELEASED_STATUS = "Fix Released"
CUSTOMER_FOUND_TAG = "customer-found"

BUGS_NEW = [NEW_STATUS, INCOMPLETE_STATUS, CONFIRMED_STATUS, TRIAGED_STATUS]
BUGS_IN_PROGRESS = [IN_PROGRESS_STATUS]
BUGS_FINISHED = [OPINION_STATUS, INVALID_STATUS, WONT_FIX_STATUS,
                 FIX_COMMITTED_STATUS, FIX_RELEASED_STATUS]

BUG_STATUSES_MAP = {
    "to-do": BUGS_NEW,
    "in-progress": BUGS_IN_PROGRESS,
    "done": BUGS_FINISHED
}


def calendar_days_ago(days):
    return datetime.datetime.now(tz=pytz.UTC) - datetime.timedelta(days=days)


def business_days_ago(days):
    today = datetime.datetime.now(tz=pytz.UTC)

    # return today if days == 0
    if days == 0:
        return today

    days_offset = 0

    # add extra days if today is weekend
    if today.isoweekday() > 5:
        days_offset += today.isoweekday() - 6

    if days <= today.isoweekday():
        days_offset += days
    else:
        days_offset += today.isoweekday()

        # add extra days for first week
        tail = (days - today.isoweekday()) % 5
        if tail > 0:
            days_offset += tail + 2

        # add extra days for whole weeks
        whole_weeks = (days - tail - today.isoweekday()) // 5
        if whole_weeks > 0:
            days_offset += whole_weeks * 7

    date = today - datetime.timedelta(days=days_offset)

    # add extra days if result date is weekend
    if date.isoweekday() > 5:
        date -= datetime.timedelta(days=2)
    return date


def check_cal_threshold(bug_date, threshold):
    return bug_date < calendar_days_ago(threshold)


def check_biz_threshold(bug_date, threshold):
    return bug_date < business_days_ago(threshold)


class BugCriteria(filters.Predicate):

    def get_hint_text(self, bug, template):
        is_customer_found = CUSTOMER_FOUND_TAG in bug.tags

        if getattr(self, 'threshold', None):
            threshold = self.threshold
        else:
            if bug.importance in [CRITICAL_IMPORTANCE, HIGH_IMPORTANCE]:
                threshold_template = bug.importance.lower()
            else:
                threshold_template = "others"
            if is_customer_found:
                threshold_template += "_customer_found"
            threshold_template += "_threshold"
            try:
                threshold = getattr(self, threshold_template)
            except AttributeError:
                if is_customer_found:
                    threshold = getattr(
                        self,
                        threshold_template.replace("_customer_found", "")
                    )

        hint_data = {
            "importance": bug.importance,
            "with_customer_found": "with `customer-found` tag" if is_customer_found else "",
            "threshold": threshold,
        }
        return template.format(**hint_data)

    @abc.abstractmethod
    def apply(self, bug):
        pass


class NonTriaged(BugCriteria):
    """Implementation for `non-triaged` criteria"""

    def __init__(self, incomplete_threshold, teams_involved, project_owners, queue_owner,
                 included_tags, involved_tags, default_queue):
        self.incomplete_threshold = incomplete_threshold
        self.teams_involved = teams_involved
        self.queue_owner = queue_owner
        self.included_tags = included_tags
        self.involved_tags = involved_tags
        self.default_queue = default_queue

        self.all_teams = db.get_major_teams()
        for k in self.all_teams.keys():
            # convert list to set to speed up search
            self.all_teams[k] = set(self.all_teams[k])

        self.owned_project = None
        for project, team in project_owners.items():
            if team == self.queue_owner:
                self.owned_project = project

    def _bug_matches_filter(self, bug):
        """This method checks if the bug relates to queue_owner or included_tags"""

        # check if bug is assigned to current queue owner
        if bug.assignee in self.all_teams[self.queue_owner]:
            return True

        # check if some other team involved in triaging bugs
        # is assigned
        for ti in self.teams_involved:
            if bug.assignee in self.all_teams[ti]:
                return False

        # check if bug doesn't contain involved tags and it is not default_queue
        other_queue_tags = set(self.involved_tags) - set(self.included_tags)
        if set(bug.tags) & other_queue_tags and self.default_queue:
            return False

        # check if bug contains included tags
        if set(bug.tags) & set(self.included_tags):
            return True

        # finally check if queue_owner is responsible for
        # bug's project (mos or fuel)
        bug_project_name = bug.target_name.split('/')[0]
        return self.owned_project == bug_project_name

    def apply(self, bug):
        if not self._bug_matches_filter(bug):
            return False

        if bug.status == INCOMPLETE_STATUS:
            return check_cal_threshold(bug.date_incomplete,
                                       self.incomplete_threshold)

        return (bug.status == NEW_STATUS or bug.milestone is None or
                bug.importance == 'Undecided' or bug.assignee is None)


class SLAFullLifecycle(BugCriteria):
    """Implementation for `sla-full-lifecycle` criteria"""

    def __init__(self, critical_threshold, high_threshold):
        self.critical_threshold = critical_threshold
        self.high_threshold = high_threshold

    def apply(self, bug):
        if bug.importance == HIGH_IMPORTANCE:
            return check_biz_threshold(bug.date_created, self.high_threshold)
        if bug.importance == CRITICAL_IMPORTANCE:
            return check_biz_threshold(bug.date_created, self.critical_threshold)
        else:
            # False for all other bugs
            return False


class SLAConfirmedTriaged(BugCriteria):
    """Implementation for `sla-confirmed-triaged` criteria"""

    def __init__(self, threshold):
        self.threshold = threshold

    def apply(self, bug):
        if bug.status in [CONFIRMED_STATUS, TRIAGED_STATUS]:
            return check_biz_threshold(bug.date_last_updated, self.threshold)
        return False


class SLAInProgress(BugCriteria):
    """Implementation for `sla-in-progress` criteria"""

    def __init__(self, critical_customer_found_threshold, critical_threshold,
                 high_customer_found_threshold, high_threshold,
                 others_threshold):
        self.critical_customer_found_threshold = critical_customer_found_threshold
        self.critical_threshold = critical_threshold
        self.high_customer_found_threshold = high_customer_found_threshold
        self.high_threshold = high_threshold
        self.others_threshold = others_threshold

    def apply(self, bug):
        if bug.status != IN_PROGRESS_STATUS:
            return False

        if bug.importance == CRITICAL_IMPORTANCE and CUSTOMER_FOUND_TAG in bug.tags:
            return check_biz_threshold(bug.date_in_progress, self.critical_customer_found_threshold)
        if bug.importance == CRITICAL_IMPORTANCE:
            return check_biz_threshold(bug.date_in_progress, self.critical_threshold)
        if bug.importance == HIGH_IMPORTANCE and CUSTOMER_FOUND_TAG in bug.tags:
            return check_biz_threshold(bug.date_in_progress, self.high_customer_found_threshold)
        if bug.importance == HIGH_IMPORTANCE:
            return check_biz_threshold(bug.date_in_progress, self.high_threshold)
        return check_biz_threshold(bug.date_in_progress, self.others_threshold)


class HCFReport(BugCriteria):
    """Implementation for `hcf-report` criteria"""

    def __init__(self, exclude_tags):
        self.exclude_tags = exclude_tags

    def apply(self, bug):
        if bool(set(self.exclude_tags) & set(bug.tags)):
            return False

        return True


class All(BugCriteria):
    """Implementation for `all` criteria"""

    def __init__(self):
        pass

    def apply(self, bug):
        return True


class IncludesTag(BugCriteria):
    def __init__(self, tag_name):
        self.tag_name = tag_name

    def apply(self, bug):
        return self.tag_name in bug.tags


class Docs(BugCriteria):
    def __init__(self, general_status):
        self.select_statuses = BUG_STATUSES_MAP[general_status]

    def apply(self, bug):
        return 'docs' in bug.tags and bug.status in self.select_statuses
