import abc
import datetime

from collections import defaultdict

import pytz
import six

from flask import current_app as app

from lpreports import db
from lpreports.launchpad.bug_task import BugTask


class Predicate(object):
    six.add_metaclass(abc.ABCMeta)

    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def apply(self, bug):
        pass


class JoinPredicate(object):

    """Join several predicates into one predicate.

    This predicate can be used for joining a number of other
    predicates with either 'and' or 'or' logic operand.
    """

    def __init__(self, predicates, join_logic='and'):
        super(JoinPredicate, self).__init__()
        self.predicates = predicates
        self.join_logic = join_logic

    def apply(self, bug):
        join_fn = all if self.join_logic == 'and' else any
        return join_fn(predicate.apply(bug) for predicate in self.predicates)


class InversePredicate(object):

    """Simple predicate which negates result of other predicate."""

    def __init__(self, predicate):
        super(InversePredicate, self).__init__()
        self.predicate = predicate

    def apply(self, bug):
        return not self.predicate.apply(bug)


class MembershipPredicate(Predicate):
    def __init__(self, bug_attr, values):
        super(MembershipPredicate, self).__init__()
        self.bug_attr = bug_attr
        self.values = values

    def apply(self, bug):
        return getattr(bug, self.bug_attr) in self.values


class TagMembershipPredicate(Predicate):
    def __init__(self, values, inclusive=True, conjunction=False):
        super(TagMembershipPredicate, self).__init__()
        self.values = values
        self.inclusive = inclusive
        self.conjunction = conjunction

    def apply(self, bug):
        if self.inclusive:
            if self.conjunction:
                # contains all tags
                return set(self.values).issubset(bug.tags)
            else:
                # contains at least one tag
                return set(self.values).intersection(bug.tags) != set()
        else:
            return set(self.values).intersection(bug.tags) == set()


class TeamMembershipPredicate(Predicate):
    def __init__(self, bug_attr, values, teams_data):
        super(TeamMembershipPredicate, self).__init__()
        if bug_attr == 'reporter':
            self.bug_attr = 'owner'
        else:
            self.bug_attr = bug_attr
        self.values = values
        self.all_people = self.flatten_teams(teams_data)
        self.teams = self.extract_real_teams(teams_data)

    def _is_unknown(self, value):
        return ('unknown' in self.values) and (value not in self.all_people)

    def _is_in_team(self, value, team, members):
        return team in self.values and (value == team or value in members)

    def apply(self, bug):
        bug_attr = getattr(bug, self.bug_attr)

        if self._is_unknown(bug_attr):
            return True

        for team, members in self.teams.items():
            if self._is_in_team(bug_attr, team, members):
                return True

        return False

    @staticmethod
    def flatten_teams(major_teams):
        flatten = set()

        for team in major_teams.values():
            for subteam, people in team.items():
                flatten.add(subteam)
                flatten = flatten.union(people)

        return flatten

    @staticmethod
    def extract_real_teams(major_teams):
        teams = {}
        for t in major_teams.values():
            teams.update(t)

        return teams

    @staticmethod
    def is_teams_filter(name):
        return name in ['reporter', 'assignee', 'verifier']


class DateRangePredicate(Predicate):
    def __init__(self, predicate_name, values):
        super(DateRangePredicate, self).__init__()
        self.attr = self.transform_predicate_name(predicate_name)
        self.postfix = predicate_name.split("_")[-1]
        self.value = self.transform_date(values)

    def apply(self, bug):
        date = getattr(bug, self.attr)

        if not date:
            return False

        date = date.replace(tzinfo=pytz.UTC)

        if self.postfix == 'to':
            return date < self.value
        elif self.postfix == 'from':
            return date > self.value
        else:
            return False

    @staticmethod
    def transform_predicate_name(predicate_name):
        return 'date_' + predicate_name. \
            replace('_to', ''). \
            replace('_from', '')

    @staticmethod
    def transform_date(date_str):
        if date_str is not None:
            return datetime.datetime.strptime(date_str, "%Y-%m-%d").replace(tzinfo=pytz.UTC)

    @staticmethod
    def is_date_filter(name):
        date_fields = ["created", "triaged", "fix_committed", "fix_released"]
        return name in ([s + "_from" for s in date_fields] +
                        [s + "_to" for s in date_fields])


class ComplexPredicate(Predicate):

    u"""Predicate which makes easier building complex filter conditions.

    Usage example::

        predicate = ComplexPredicate({
            'tags': ['feature', 'tech-debt']
            'importance': ['Low', 'Medium', 'Undecided' ]
        })

        This `predicate` object  will give `True` only for bugs
        satisfying two conditions:
            1. Bug is marked with 'feature' or 'tech-debt' tag.
            2. Bug has 'Low', 'Medium' or 'Undecided' importance.
    """

    def __init__(self, rules):
        super(ComplexPredicate, self).__init__()
        self.predicate = self._make_predicate(rules)

    def apply(self, bug):
        return self.predicate.apply(bug)

    @classmethod
    def _make_predicate(cls, rules):
        predicates = []

        for attr, values in rules.iteritems():

            if attr == 'tags':
                pred = TagMembershipPredicate(values)
            else:
                pred = MembershipPredicate(attr, values)

            predicates.append(pred)
        return JoinPredicate(predicates)


def compile_predicates(params, teams_data):
    predicates = []
    teams_predicates = []
    teams_and_area_tags = False

    for name, values in params.items():
        pred = None
        teams_pred = None

        if DateRangePredicate.is_date_filter(name):
            pred = DateRangePredicate(name, values)
        elif TeamMembershipPredicate.is_teams_filter(name):
            teams_pred = TeamMembershipPredicate(name, values, teams_data)
        elif name == 'area_tags':
            area_tags = []
            for val in values:
                tags = db.get_area_tags(val)
                if tags:
                    area_tags.extend(tags)
            teams_pred = TagMembershipPredicate(area_tags, inclusive=True)
        elif name == 'tags':
            pred = TagMembershipPredicate(values, inclusive=True)
        elif name == 'tags_all':
            pred = TagMembershipPredicate(values, inclusive=True,
                                          conjunction=True)
        elif name == 'exclude_tags':
            pred = TagMembershipPredicate(values, inclusive=False)
        elif name == 'exclude_tags_all':
            pred = TagMembershipPredicate(values, inclusive=False,
                                          conjunction=True)
        elif name == 'component':
            pass
        elif name == 'original_milestone':
            pass
        elif name == 'assigned_person':
            pred = MembershipPredicate('assignee', values)
        elif name == 'teams_and_area_tags':
            teams_and_area_tags = True if values else False
        elif name == 'only_user_affecting':
            excludes_conf = app.config['REPORT_CONFIG'][
                'not-user-affecting-bugs']
            pred = JoinPredicate(
                [InversePredicate(ComplexPredicate(c)) for c in excludes_conf])
        elif values:
            pred = MembershipPredicate(name, values)

        if pred:
            predicates.append(pred)
        if teams_pred:
            teams_predicates.append(teams_pred)

    if teams_predicates:
        join_logic = 'and' if teams_and_area_tags else 'or'
        predicates.append(JoinPredicate(teams_predicates, join_logic=join_logic))

    return JoinPredicate(predicates).apply


def prepare_query(query):
    # get rid of all the params with empty predicates
    return dict([i for i in query.items() if i[1]])


def cut_milestone_version(milestone):
    """This function cuts milestone suffix and returns tuple of numbers.

    For example: incoming arg = 9.0-updates, value which returns: (9, 0)
    """
    if not milestone:
        return tuple()

    # set max weight to 'next' milestone
    if milestone == 'next':
        return 99999,

    num_part = milestone.split('-')[0]
    return tuple(num_part.split('.'))


def make_id2ms_map(bugs):
    """The function makes id > milestone map for each bug."""
    id2ms = defaultdict(set)
    for bug in bugs:
        id2ms[bug.id].add(cut_milestone_version(bug.milestone))
    return id2ms


def origin_milestone_filter(bugs, original_milestones):
    """This function removes all bugs which haven't been created in original milestones."""
    # remove suffixes from each milestone and convert it to set
    original_milestones = {cut_milestone_version(m) for m in original_milestones}
    # make id2ms map
    id2ms_map = make_id2ms_map(bugs)

    filtered_bugs = list()
    for bug in bugs:
        # skip bug if it doesn't contain at least one of origin milestones
        if not original_milestones & id2ms_map[bug.id]:
            continue
        # add bugs to result if it doesn't contain milestones which
        # less then min original milestone
        if min(original_milestones) <= min(id2ms_map[bug.id]):
            filtered_bugs.append(bug)

    return filtered_bugs


def group_bugs_by_id(bugs):
    """The function groups bugs by id and returns following dict {bug_id: [bug0, bug1 ...]}"""
    id2bugs = defaultdict(list)
    for bug in bugs:
        id2bugs[bug.id].append(bug)
    return id2bugs


def cmp_by_milestone(x, y):
    """Custom comparison function to sort bug_task by milestone

    The function uses default comparison mechanism from Milestone class but
    places version that doesn't have suffix into the top of list.
    """
    if x.milestone and not y.milestone:
        return -1
    elif not x.milestone and y.milestone:
        return 1
    elif not x.milestone and not y.milestone:
        return 0

    if x.milestone.suffix and not y.milestone.suffix:
        return 1
    elif not x.milestone.suffix and y.milestone.suffix:
        return -1
    if x.milestone > y.milestone:
        return -1
    elif x.milestone < y.milestone:
        return 1
    return 0


def merge_bug_tasks(bugs, id2bugs):
    """Bug task merge functions

    This functions finds bugs with the same id and merge them into single
    bug with some bug_tasks
    """
    merged_bugs = {}

    for bug in bugs:
        # if was found more than one bug with one id.
        if len(id2bugs[bug.id]) > 1:
            # skip bug if it has been already added
            if merged_bugs.get(bug.id):
                continue
            # convert bug into bug_tasks and sort them by milestone
            bug.bug_tasks = sorted(
                map(BugTask, id2bugs[bug.id]),
                cmp=cmp_by_milestone
            )
            # set max weight to main bug
            bug.weight = max([b.weight for b in id2bugs[bug.id]])

        merged_bugs[bug.id] = bug

    return merged_bugs.values()


def filter_bugs(bugs, search_params, teams_data):
    params = prepare_query(search_params)

    # filter by original milestone
    original_milestone = search_params.get('original_milestone')
    if original_milestone:
        bugs['DATA'][0]['bugs'] = origin_milestone_filter(
            bugs['DATA'][0]['bugs'],
            original_milestone
        )

    predicates = compile_predicates(params, teams_data)

    bugs['DATA'][0]['bugs'] = filter(predicates, bugs['DATA'][0]['bugs'])

    # group by teams
    if len(bugs['DATA']) == 2 and bugs['DATA'][1]:
        teams = bugs['DATA'][1]

        for bug in bugs['DATA'][0]['bugs']:
            teams[bug.major_team]['bugs'].append(bug)

        bugs['DATA'] = teams.values()

    # merge bug_tasks
    for data in bugs['DATA']:
        if data.get('bugs'):
            id2bugs = group_bugs_by_id(data['bugs'])
            data['bugs'] = merge_bug_tasks(data['bugs'], id2bugs)

    return bugs
