import datetime

from bisect import bisect_left
from collections import defaultdict
from collections import OrderedDict

import pytz

from lpreports import db
from lpreports.launchpad import bug as b
from lpreports.utils import common as utils


class ReleaseChart(object):

    def __init__(self, project_name, milestone_names):
        self.bugs = []
        for status in b.BUG_STATUSES.values():
            self.bugs += db.get_bugs(project_name,
                                     status,
                                     milestone_names)

        self.initial_date = datetime.datetime.now() - datetime.timedelta(days=120)

    def get_trends_data(self):

        # the chart will span until tomorrow
        window_end = datetime.datetime.now(pytz.utc) + datetime.timedelta(
            days=1)

        # add chart series in order from bottom to top
        data = OrderedDict()
        data["Verified"] = []
        data["Resolved"] = []
        data["In Progress"] = []
        data["Open"] = []
        data["Incomplete"] = []

        # all dates
        all_dates = set()

        # process each bug and its events
        for _bug in self.bugs:
            events = _bug.get_status_changes()
            events.append({"date": window_end, "type": "N/A"})

            for i in range(0, len(events) - 1):
                e1 = events[i]
                e2 = events[i + 1]

                t = e1["type"]
                d1 = e1["date"].replace(tzinfo=None)
                d2 = e2["date"].replace(tzinfo=None)

                if d1 <= d2:
                    d1 = d1.replace(hour=0, minute=0, second=0, microsecond=0)
                    d2 = d2.replace(hour=0, minute=0, second=0, microsecond=0)

                    data[t].append({"date": d1, "num": 1})
                    data[t].append({"date": d2, "num": -1})
                    all_dates.add(d1)
                    all_dates.add(d2)

        # create the list of all sorted dates
        all_dates_sorted = sorted(all_dates)
        n = len(all_dates_sorted)

        # process each data item and construct chart
        d3_start = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, pytz.utc)

        chart = []
        for t in data:
            events = sorted(data[t], key=lambda d: (d['date'], -d['num']))

            # for each date, mark result in global list with dates
            all_dates_values = [None] * n
            bug_count = 0
            for e in events:
                bug_count += e["num"]
                idx = bisect_left(all_dates_sorted, e["date"])
                if not all_dates_sorted[idx] == e["date"]:
                    raise ValueError(
                        "Date not found in array using binary search")
                all_dates_values[idx] = bug_count

            # process all global dates
            prev = 0
            for idx in range(0, n):
                if all_dates_values[idx] is not None:
                    prev = all_dates_values[idx]
                    break

            for idx in range(0, n):
                if all_dates_values[idx] is None:
                    all_dates_values[idx] = prev
                else:
                    prev = all_dates_values[idx]

            # create series for the chart
            # (except for the last point, which has all zeroes)
            values = []
            for idx in range(0, n - 1):
                if all_dates_sorted[idx] < self.initial_date:
                    continue

                chart_seconds = (all_dates_sorted[idx].replace(
                    tzinfo=None) - d3_start.replace(
                    tzinfo=None)).total_seconds() * 1000.0
                values.append([int(chart_seconds), all_dates_values[idx]])
            chart.append({'key': t, 'values': values})

        return chart

    def get_incoming_outgoing_data(self, tickday=utils.Weekday.MONDAY):
        # add chart series in order from bottom to top
        data = OrderedDict()
        data["Incoming"] = defaultdict(int)
        data["Outgoing"] = defaultdict(int)

        # all dates
        all_dates = set()

        # process each bug and its events
        for _bug in self.bugs:
            for e in _bug.get_status_changes():
                # get the date for a given day
                bug_weekday = datetime.timedelta(days=e["date"].weekday())
                report_tickday = datetime.timedelta(days=tickday)
                week_corr = datetime.timedelta(
                    # events before the 'tick' day are counted in prev week
                    weeks=1 if e["date"].weekday() < tickday else 0
                )
                date = e["date"] - bug_weekday + report_tickday - week_corr

                date = date.replace(hour=0, minute=0, second=0, microsecond=0)

                # process incoming and outgoing bugs
                if e["type"] == "Open":
                    data["Incoming"][date] += 1
                    all_dates.add(date.replace(tzinfo=pytz.UTC))
                elif e["type"] == "Resolved":
                    data["Outgoing"][date] += 1
                    all_dates.add(date.replace(tzinfo=pytz.UTC))

        # create the list of all sorted dates
        all_dates_sorted = sorted(all_dates)
        initial_date = self.initial_date.replace(tzinfo=pytz.UTC)
        all_dates_sorted = filter(lambda dt: dt >= initial_date, all_dates_sorted)
        n = len(all_dates_sorted)

        # process each data item and construct chart
        d3_start = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, pytz.utc)

        chart = []
        for t in data:
            # for each date, mark result in global list with dates
            all_dates_values = [None] * n
            for idx in range(0, n):
                date = all_dates_sorted[idx]
                all_dates_values[idx] = data[t][date]

            # create series for the chart
            values = []
            for idx in range(0, n):
                chart_seconds = (all_dates_sorted[idx].replace(
                    tzinfo=None) - d3_start.replace(
                    tzinfo=None)).total_seconds() * 1000.0
                values.append([int(chart_seconds), all_dates_values[idx]])
            chart.append({'key': t, 'values': values})

        return chart
