from collections import OrderedDict
from itertools import izip

from flask import current_app as app

from lpreports import db
from lpreports.launchpad.bug import Bug
from lpreports.launchpad import criterias as criterias_implementation

BUG_IMPORTANCE = ["Essential", "Critical", "High", "Medium",
                  "Low", "Wishlist", "Unknown", "Undecided"]

CUSTOMER_FOUND_TAG = "customer-found"


def get_criteria_by_name(criteria_name):
    conf_criterias = app.config["REPORT_CONFIG"]['criterias']
    if criteria_name not in [_['name'] for _ in conf_criterias]:
        raise ValueError("There is no %r section in 'criterias' config "
                         "section" % criteria_name)

    return [cr for cr in conf_criterias if cr['name'] == criteria_name][0]


def get_criteria_implementation(config_criteria, milestone_names):
    """Get implementation of the criteria, based on config.

    This function checks for config criteria_name in config['criterias'],
    gets class - realization of this criteria - and returns an instance
    of this class.
    """

    criteria = get_criteria_by_name(config_criteria['name'])
    class_name = criteria['implementation'].split('.')[-1]
    try:
        ImplClass = getattr(criterias_implementation, class_name)
    except AttributeError:
        raise ValueError("There is no implementation for %r class in "
                         "'criterias' file " % class_name)

    overrides = config_criteria.get('config-override', {})

    kwargs = {}
    for conf in criteria.get('config', []):
        cname = conf['name']

        override_name = "{0}_{1}".format(', '.join(milestone_names), cname)
        if override_name in overrides:
            kwargs[cname] = overrides[override_name]
        else:
            kwargs[cname] = overrides.get(cname, conf['default'])

    return ImplClass(**kwargs)


def get_criteria_description(criterias, milestone_names):
    """Get implementation of the criteria, based on config.

    This function checks for config criteria_name in config['criterias'],
    gets class - realization of this criteria - and returns an instance
    of this class.
    """

    description = ""
    for config_criteria in criterias:
        criteria = get_criteria_by_name(config_criteria['name'])

        if 'config-override' not in config_criteria:
            kwargs = {conf.get("text", conf['name']): conf['default']
                      for conf in criteria.get('config', [])}
        else:
            kwargs = {}
            for conf in criteria.get('config', []):
                override_name = "{0}_{1}".format(', '.join(milestone_names),
                                                 conf['name'])
                detailed_name = conf.get("text", conf['name'])
                if override_name in config_criteria['config-override']:
                    kwargs[detailed_name] = \
                        config_criteria['config-override'][override_name]
                else:
                    kwargs[detailed_name] = conf['default']

        criteria_name = criteria.get('text', criteria['name'])
        description += "%s (%s):\n" % (criteria_name, criteria['name'])
        description += "\n".join("\t%s: %s" % (k, v) for (k, v) in kwargs.items())
        description += "\n"

    return description.strip()


def should_show_bug(bug, user_security_membership):
    return bug.get("information_type", None) != "Private Security" or (
        bug["information_type"] == "Private Security" and
        bug["target_name"] in user_security_membership)


def get_bugs_by_criteria(criterias, projects, milestone_names, options={},
                         security_membership=[], skip_duplicates=True):
    # Create filters, based on incoming data, to get bugs from db
    filters = list()
    filters.append({"status": {"$in": options["status"]}})
    filters.append({"importance": {"$in": options["importance"]}})

    if milestone_names:
        filters.append({"milestone": {"$in": milestone_names}})

    if skip_duplicates:
        filters.append({"is_duplicate": {"$ne": True}})

    filtered_bugs = list()
    for project in projects:
        for bug in db.bugs_db()[project].find({"$and": filters}):
            if should_show_bug(bug, security_membership):
                filtered_bugs.append(Bug(bug))

    matching_bugs = OrderedDict()
    for criteria in criterias:

        implementation = get_criteria_implementation(criteria, milestone_names)
        criteria = get_criteria_by_name(criteria['name'])

        for bug, object_id in izip(filtered_bugs, (id(b) for b in filtered_bugs)):
            if implementation.apply(bug):
                try:
                    hint_text = implementation.get_hint_text(
                        bug, criteria.get('hint-text', ''))
                except (UnboundLocalError, AttributeError):
                    # something wrong with hint text
                    hint_text = ''
                else:
                    hint_text += ' ({0})'.format(criteria['name'])

                if matching_bugs.get(object_id):
                    matching_bugs[object_id].criteria += "\n{0}".format(criteria['name'])
                    matching_bugs[object_id].criteria_hint_text += "\n{0}".format(hint_text)
                else:
                    bug.criteria = criteria['name']
                    bug.criteria_hint_text = hint_text
                    matching_bugs[object_id] = bug

    return list(matching_bugs.itervalues())


def get_reports_data(report, projects,
                     milestone_names=[], security_membership=[]):
    bugs = get_bugs_by_criteria(
        criterias=report['criterias'],
        projects=projects,
        milestone_names=milestone_names,
        options=report['options'],
        security_membership=security_membership)
    report_legend = get_criteria_description(report['criterias'],
                                             milestone_names)

    bugs = {
        'name': report['name'],
        'display_name': report['text'],
        'parameter': report['parameter'],
        'bugs': bugs,
        'report_legend': report_legend,
    }

    teams_info = OrderedDict()
    if report.get('group-by') == "team":
        for team in db.get_major_teams().keys():
            info = {
                'name': "%s for %s team" % (report['name'], team),
                'display_name': team,
                'parameter': report['parameter'],
                'bugs': [],
                'report_legend': report_legend,
            }
            teams_info[team] = info

    return {"DATA": [bugs, teams_info]}
