from functools import wraps

from flask import current_app
from flask import g
from flask import request
from flask_restful import abort
from flask_restful import Resource
from itsdangerous import BadData
from itsdangerous import JSONWebSignatureSerializer

from lpreports import db


def _is_auth_token_valid(token):
    serializer = JSONWebSignatureSerializer(
        current_app.config["SECRET_KEY"])

    try:
        user_email = serializer.loads(token)
    except BadData:
        return False

    user = db.find_user(email=user_email)
    if user:
        g.user = user
        return True

    return False


def token_auth_required(f):
    """Ensures that request has "Authorization" header with valid auth token.

    HTTP header example:
        Authorization: Token dyJhbGciOiJUUzI1NiJ9.ImZraGF0YW1vdkBta
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        auth_header = request.headers.get('Authorization')
        if auth_header is None:
            abort(401, message='Unauthorized access')

        marker, token = (auth_header.split(' ', 1) + [''])[:2]
        if marker.lower() == 'token' and _is_auth_token_valid(token):
            return f(*args, **kwargs)

        abort(401, message='Invalid auth token')

    return decorated


class BaseResource(Resource):

    """Base resource for Rest API.

    All methods require token authentication. Token can be obtained
    by requesting /api/v1.0/auth_token resource.
    """

    method_decorators = [token_auth_required]


class AuthTokenResource(Resource):

    def get(self):
        serializer = JSONWebSignatureSerializer(
            current_app.config["SECRET_KEY"])
        return {'token': serializer.dumps(g.user['email'])}


class ErrorResource(BaseResource):

    def get(self, path='/'):
        abort(404, message='Path does not exist')
