from datetime import datetime
import re

from flask import current_app as app

from lpreports import db


def process_maintenance_bugs(bugs):
    nearest_milestone = get_nearest_targeted_milestone()

    maint_inventory = MaintenanceRulesEngine()

    for data in bugs['DATA']:
        if data.get('bugs'):
            for bug in data['bugs']:
                maint_inventory.check(bug, nearest_milestone)

    return bugs


def sort_maintenance_bugs(bugs):
    for data in bugs['DATA']:
        if data.get('bugs'):
            data['bugs'] = sorted(
                data['bugs'],
                key=lambda bug: bug.weight,
                reverse=True)
    return bugs


def get_nearest_targeted_milestone():
    today = datetime.today()
    nearest_milestone = None
    target_dates = db.get_milestones_target_dates().items()
    if target_dates:
        nearest_milestone, nearest_target_date = target_dates[0]
        nearest_target_date = nearest_target_date.replace(tzinfo=None)
        least_delta = nearest_target_date - today
        for milestone, target_date in target_dates[1:]:
            target_date = target_date.replace(tzinfo=None)
            delta = target_date - today
            if delta < least_delta and delta.days >= 0:
                least_delta = delta
                nearest_milestone = milestone

    return nearest_milestone


class MaintenanceRulesEngine(object):
    _rules = []
    versions_mapping = {}
    maintenance_series = {}

    def __init__(self):
        if not self._rules:
            for rule in app.config['REPORT_CONFIG']['maintenance_bug_rules']:
                self._rules.append(self.compile_rule(rule))
        if not self.versions_mapping:
            self.versions_mapping = app.config[
                'REPORT_CONFIG']['maintenance_ct_versions_mapping']
        if not self.maintenance_series:
            self.maintenance_series = app.config[
                'REPORT_CONFIG']['maintenance_series']

    def check(self, bug, nearest_milestone):
        weight = 0
        legend = []
        for rule in self._rules:
            result = rule(bug)
            if result:
                if 'weight-strategy' in result:
                    ws_handler = getattr(
                        self, "weight_strategy_" +
                        result['weight-strategy'].replace(
                            '-', '_'))
                    result = ws_handler(bug)
                if result:
                    weight += result['weight']
                    legend.append((result['weight'], result['legend']))
        if nearest_milestone and bug.milestone == nearest_milestone:
            weight += 100
            legend.append((100, "within nearest milestone"))
        if weight:
            bug.set_weight(weight)
            bug.set_legend(legend)

    def compile_rule(self, rule):
        def compiled_rule(bug):
            result = all([getattr(self, "check_" + criterion.replace('-', '_'))(bug, value)
                         for criterion, value in rule['criteria'].items()])
            if result:
                if 'weight-strategy' in rule:
                    return {'weight-strategy': rule['weight-strategy']}
                return {'weight': rule['weight'],
                        'legend': rule['legend']}
            else:
                return None
        return compiled_rule

    @staticmethod
    def check_tags(bug, rule_tags):
        return any([tag in rule_tags for tag in bug.tags])

    @staticmethod
    def check_status(bug, rule_statuses):
        return any([status in bug.status for status in rule_statuses])

    @staticmethod
    def check_milestone_globs(bug, milestone_globs):
        return bug.milestone and any([milestone_glob in bug.milestone
                                      for milestone_glob in milestone_globs])

    @staticmethod
    def check_importance(bug, importance):
        return bug.importance == importance

    @staticmethod
    def check_importance_inverted(bug, importance):
        return bug.importance != importance

    @staticmethod
    def check_security(bug, is_security):
        result = 'Security' in bug.information_type
        return (result if is_security else not result)

    def weight_strategy_cumulative(self, bug):
        series = self.maintenance_series.get(bug.target_name)
        if not series:
            return False
        cumulative_score = 0
        for bug_task in bug.bug_tasks:
            for s in series:
                if s in str(bug_task.milestone):
                    cumulative_score += 1
                    break
        return {'weight': cumulative_score,
                'legend': 'Nominated to {0} adjacent series'.format(
                    cumulative_score)} if cumulative_score > 1 else False

    def check_tags_re(self, bug, tags_re_dict):
        if not all([key in tags_re_dict for key in ['group', 're', 'handler']]):
            raise ValueError("Rule error in 'tags_re', check config")
        else:
            for tag in bug.tags:
                match = re.search(tags_re_dict['re'], tag)
                if match:
                    result = match.groups(int(tags_re_dict['group']))
                    return getattr(self, "re_handler_" + tags_re_dict[
                        'handler'].replace('-', '_'))(bug, result)
            return False

    def re_handler_cl_patched_tags(self, bug, value):
        return bug.milestone == self.versions_mapping[value]
