import cStringIO
import csv

import flask
from flask import current_app as app
from flask import g

from lpreports import db
from lpreports.launchpad import bug as lp_bug
from lpreports.launchpad import sla_reports
from lpreports import util
from lpreports.utils.maintenance import process_maintenance_bugs
from lpreports.utils.maintenance import sort_maintenance_bugs

reports = flask.Blueprint('reports', __name__)


def _make_csv(bugs):
    header = lp_bug.FIELDS_TO_COPY[:]
    header.append("team")

    rows = [header]
    for team_data in bugs['DATA']:
        for bug in team_data.get('bugs', []):
            # There are non-ascii characters in bug titles
            # and we need to handle that explicitly
            title = bug.title.encode('ascii', 'ignore')

            row = []
            for field in lp_bug.FIELDS_TO_COPY:
                if field == "tags":
                    tags = getattr(bug, "tags") or []
                    row.append(" ".join(tags))
                elif field == "title":
                    row.append(title)
                else:
                    row.append(getattr(bug, field))

            row.append(team_data["display_name"])
            rows.append(row)

    str_io = cStringIO.StringIO()
    csvwriter = csv.writer(str_io)
    csvwriter.writerows(rows)

    return flask.Response(str_io.getvalue(),
                          mimetype='text/plain',
                          headers={"Content-Disposition":
                                   "attachment;filename=report.csv"})


def count_unique_bugs(bugs):
    unique_ids = set()
    for group in bugs['DATA']:
        for bug in group.get('bugs', []):
            unique_ids.add(bug.id)
    return len(unique_ids)


def _render_custom_report(report):
    default_projects = app.config["DEFAULT_PROJECTS"]
    all_projects = app.config["PROJECTS_LIST"]
    projects = flask.request.args.getlist('project') or default_projects

    user = g.user
    security_membership = db.get_user_security_teams_membership(user["email"])

    bugs = sla_reports.get_reports_data(
        report=report,
        projects=projects,
        security_membership=security_membership
    )
    teams_data = db.get_full_teams()

    actual_milestones = db.get_milestones()

    bugs, filters = util.filter_bugs(flask.request, bugs, teams_data)

    if report['name'] == 'maintenance_report':
        actual_milestones = db.get_maintenance_milestones()
        bugs = sort_maintenance_bugs(process_maintenance_bugs(bugs))
    else:
        # NOTE(aermolov): Bugs should be sorted by priority and "customer-found" tag
        # in bug tags.
        for data in bugs['DATA']:
            if data.get('bugs'):
                data['bugs'] = sorted(data['bugs'], key=lambda bug: (
                    sla_reports.BUG_IMPORTANCE.index(bug.importance),
                    sla_reports.CUSTOMER_FOUND_TAG not in bug.tags))
    milestone_field = True if report['parameter'] == 'milestone' else False

    color_map = {
        'red': 'label-danger',
        'yellow': 'label-warning',
        'green': 'label-success',
        'cyan': 'label-info',
        'blue': 'label-primary',
    }
    label_types = {
        k: color_map[v['color']] for k, v in app.config['REPORT_CONFIG']['tag-labels'].items()
    }

    if flask.request.args.get('format', 'html') == 'csv':
        return _make_csv(bugs)

    return flask.render_template(
        "report.html",
        report=report,
        all_bugs=bugs,
        unique_bugs_count=count_unique_bugs(bugs),
        teams=db.get_teams(),
        area_tags=db.get_area_tag_teams(),
        filters=filters,
        all_projects=all_projects,
        projects=projects,
        milestone_field=milestone_field,
        milestones=actual_milestones,
        label_types=label_types
    )


@reports.route("/<report_name>")
def show(report_name):
    report = util.get_report_by_name(report_name)
    if not report:
        return flask.abort(404)

    return _render_custom_report(report)
